# -*- coding: utf-8 -*-
from pathlib import Path

from vise.chempotdiag.gas import Gas

from vise.config import ROOM_TEMPERATURE, REFERENCE_PRESSURE
from vise.util.testing import ViseTest

molecules_dir = Path(__file__).parent / ".." / "molecules"


class TestGas(ViseTest):

    def setUp(self):
        self.o2 = Gas("O2", temperature=300)
        self.nh3 = Gas("NH3", temperature=1000)

    def test_o2 (self):
        self.assertAlmostEqual(self.o2.trans_free_energy, -0.40862436734991214)
        self.assertAlmostEqual(self.o2.rot_free_energy, -0.1107263561129291)
        self.assertAlmostEqual(self.o2.zero_point_vibrational_energy, 0.09720351919699763)
        self.assertAlmostEqual(self.o2.vib_free_energy, -1.401901140122062e-05)
        self.assertAlmostEqual(self.o2.spin_free_energy, -0.02840132465202343)
        self.assertAlmostEqual(self.o2.energy_shift, -0.2252812739646341)

    def test_nh3(self):
        self.assertAlmostEqual(self.nh3.rot_free_energy, -0.5283431627639921)

    def test_assert_error(self):
        with self.assertRaises(ValueError):
            low_temp_nh3 = Gas("NH3", temperature=100)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pymatgen.electronic_structure.core import Spin
from pathlib import Path
import tempfile
import unittest
import warnings

from vise.analyzer.dos_plotter import get_dos_plot, max_density
from vise.util.testing import ViseTest

parent_dir = Path(__file__).parent


class ViseDosPlotterNormalTest(ViseTest):
    @classmethod
    def setUpClass(cls):
        warnings.simplefilter("ignore")

    @classmethod
    def tearDownClass(cls):
        warnings.simplefilter("default")

    @unittest.skipIf(not ViseTest.DISPLAY_DIAGRAM, ViseTest.no_display_reason)
    def test(self):
        with tempfile.TemporaryDirectory() as tmp_dirname:
            dos_vasprun = parent_dir / "MgO_dos_vasprun.xml"
            self.mgo_dos = get_dos_plot(vasprun=dos_vasprun)
            self.mgo_dos.show()
            self.mgo_dos.savefig(Path(tmp_dirname) / "MgO_dos.pdf")


class ViseDosPlotterWithArgsTest(ViseTest):
    @classmethod
    def setUpClass(cls):
        warnings.simplefilter("ignore")

    @classmethod
    def tearDownClass(cls):
        warnings.simplefilter("default")

    @unittest.skipIf(not ViseTest.DISPLAY_DIAGRAM, ViseTest.no_display_reason)
    def test(self):
        nan3_dos_mod = get_dos_plot(vasprun=parent_dir / "NaN3_dos_vasprun.xml",
                                    pdos_type="none",
                                    orbital=False,
                                    xlim=[-3, 5],
                                    ymaxs=[10, 5],
                                    zero_at_efermi=False,
                                    specific=["1"],
                                    crop_first_value=False,
                                    show_spg=False)
        nan3_dos_mod.show()
        with tempfile.TemporaryDirectory() as tmp_dirname:
            nan3_dos_mod.savefig(Path(tmp_dirname) / "NaN3_dos.pdf")


class ViseDosPlotterFerromagneticTest(ViseTest):
    @classmethod
    def setUpClass(cls):
        warnings.simplefilter("ignore")

    @classmethod
    def tearDownClass(cls):
        warnings.simplefilter("default")

    @unittest.skipIf(not ViseTest.DISPLAY_DIAGRAM, ViseTest.no_display_reason)
    def test(self):
        mno_dos = get_dos_plot(vasprun=parent_dir / "MnO_uniform_vasprun.xml")
        mno_dos.show()
        with tempfile.TemporaryDirectory() as tmp_dirname:
            mno_dos.savefig(Path(tmp_dirname) / "MnO_dos.pdf")


class MaxDensityTest(ViseTest):
    def test(self):
        density = {Spin.up: [10.0, 1.0, 2.0, 3.0],
                   Spin.down: [4.0, 8.0, 6.0, 7.0]}
        energies = [0.0, 10.0, 20.0, 230.0]
        xlim = [5.0, 25.0]
        crop_first_value = True

        expected = 8.0
        actual = max_density(density,
                             energies,
                             xlim,
                             crop_first_value)
        self.assertEqual(expected, actual)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pathlib import Path
import unittest
import warnings

from vise.util.testing import ViseTest
from vise.analyzer.band_plotter import (
    PrettyBSPlotter, ModBSPlotter, VaspBandStructureSymmLine)

parent_dir = Path(__file__).parent


class PlotTwoBandStructureTest(ViseTest):
    @classmethod
    def setUpClass(cls):
        warnings.simplefilter("ignore")

    @classmethod
    def tearDownClass(cls):
        warnings.simplefilter("default")

    @unittest.skipIf(not ViseTest.DISPLAY_DIAGRAM, ViseTest.no_display_reason)
    def test(self):
        mgo_band = VaspBandStructureSymmLine(
            parent_dir / "MgO_band_KPOINTS",
            parent_dir / "MgO_band_vasprun.xml", is_projection=True)
        mgo_nhse_band = \
            VaspBandStructureSymmLine(parent_dir / "MgO_band_KPOINTS",
                                      parent_dir / "MgO_band_nhse_vasprun.xml")
        plotter = PrettyBSPlotter(band=mgo_band, band2=mgo_nhse_band,
                                  absolute=False)
        plotter.show()


class PlotFerromagneticBandStructureTest(ViseTest):
    @classmethod
    def setUpClass(cls):
        warnings.simplefilter("ignore")

    @classmethod
    def tearDownClass(cls):
        warnings.simplefilter("default")

    @unittest.skipIf(not ViseTest.DISPLAY_DIAGRAM, ViseTest.no_display_reason)
    def test(self):
        ko2_band = VaspBandStructureSymmLine(
            kpoints_filename=(parent_dir / "KO2_band_KPOINTS"),
            vasprun_filename=(parent_dir / "KO2_band_vasprun.xml"),
            is_projection=True)
        plotter = PrettyBSPlotter(band=ko2_band, absolute=False)
        plotter.show()




# -*- coding: utf-8 -*-

import shutil
from pathlib import Path
from typing import List, Optional, Union
import yaml

from pymatgen import Element, MPRester, Composition

from vise.util.logger import get_logger
from vise.chempotdiag.gas import MOLECULE_DATA

logger = get_logger(__name__)


def get_mp_materials(elements: List[str],
                     properties: List[str],
                     e_above_hull: float = 1e-4,
                     api_key: str = None) -> list:
    """Get Materials Project materials composed of the given elements as list

    Args:
        elements (list):
            Element list like ["Cu", "O"]
        properties:
            Returned properties.
        e_above_hull:
            Energy criterion in eV/atom.
        api_key:
            Materials Project REST api key.

    Returns:
        List of materials composed of properties.
    """
    exclude_z = list(i for i in range(1, 100))
    excluded_elements = [str(Element.from_Z(z)) for z in exclude_z]
    for e in elements:
        excluded_elements.remove(e)
    with MPRester(api_key) as m:
        materials = \
            m.query(criteria={"elements": {"$in": elements,
                                           "$nin": excluded_elements},
                              "e_above_hull": {"$lte": e_above_hull}},
                    properties=properties)

    return materials


def make_poscars_from_mp(elements: list,
                         path: Union[Path, str] = Path.cwd(),
                         e_above_hull: float = 0.01,
                         api_key: str = None,
                         molecules: bool = True,
                         properties: Optional[list] = None,
                         only_unary=False) -> None:
    """Create POSCAR files composed of the given elements

    The prior_info.yaml files that are used for generating VASP input files are
    also generated by using the materials project database.

    Args:
        elements(list):
        e_above_hull(float):
        api_key(str):
        properties (list):
            See docstrings of get_mp_materials above.
        path(str):
            Path to create directories.
        molecules(bool):
            Whether to replace molecular entries to molecules in a large box.
        only_unary:
            Create only unary simple substances.

    Returns:
        None
    """
    path = Path(path)
    if not path.is_dir:
        raise NotADirectoryError(f"{path} is not directory.")

    mol_dir = Path(__file__).parent / ".." / "chempotdiag" / "molecules"

    molecules_formula_list = []
    if molecules:
        for g in MOLECULE_DATA.keys():
            comp = Composition(str(g))
            if set([str(e) for e in comp.elements]) < set(elements):
                molecules_formula_list.append(comp.reduced_formula)
                dirname = path / f"mol_{str(comp)}"
                if dirname.exists():
                    logger.error(f"{dirname} exists! So, skip creating it.")
                else:
                    dirname.mkdir()
                    shutil.copyfile(mol_dir / str(comp) / "POSCAR",
                                    dirname / "POSCAR")
                    shutil.copyfile(mol_dir / str(comp) / "prior_info.yaml",
                                    dirname / "prior_info.yaml")

    properties = properties or ["task_id",
                                "full_formula",
                                "final_energy",
                                "structure",
                                "spacegroup",
                                "band_gap",
                                "total_magnetization",
                                "magnetic_type"]
    if only_unary:
        with MPRester(api_key) as m:
            materials = \
                m.query(criteria={"elements": {"$in": elements},
                                  "nelements": 1,
                                  "e_above_hull": {"$lte": e_above_hull}},
                        properties=properties)
    else:
        materials = \
            get_mp_materials(elements, properties, e_above_hull, api_key)

    for m in materials:
        comp = Composition(m.pop("full_formula"))
        formula = comp.reduced_formula
        if molecules and formula in molecules_formula_list:
            continue

        m_path = path / f"{formula}_{m['task_id']}"
        m_path.mkdir()
        m.pop("structure").to(filename=m_path / "POSCAR")
        yaml_path = m_path / "prior_info.yaml"
        with open(str(yaml_path), "w") as fw:
            fw.write(yaml.dump(m))



# -*- coding: utf-8 -*-

from glob import glob
import tempfile
import unittest
from unittest.mock import patch

from vise.util.mp_tools import get_mp_materials, make_poscars_from_mp
from vise.util.testing import ViseTest


class TestGetMpMaterials(ViseTest):

    @unittest.skipIf(not ViseTest.PMG_MAPI_KEY, ViseTest.no_mapi_key)
    def test_mprester(self) -> None:
        materials = get_mp_materials(elements=["Si"],
                                     properties=["band_gap"],
                                     e_above_hull=0.000001)
        print(materials)
        self.assertAlmostEqual(0.6244, materials[0]["band_gap"])

    @patch("pymatgen.MPRester.query")
    def test_with_mock(self, mock) -> None:
        materials = get_mp_materials(elements=["Si"],
                                     properties=["band_gap"],
                                     e_above_hull=0.000001)
        self.assertEqual(mock(), materials)
        mock.assert_called()


class TestMakePoscarsFromMp(ViseTest):
    @unittest.skipIf(not ViseTest.PMG_MAPI_KEY, ViseTest.no_mapi_key)
    def test_get_poscars(self):
        with tempfile.TemporaryDirectory() as tmp_dirname:
            make_poscars_from_mp(elements=["Mg", "O"],
                                 path=tmp_dirname)
            files = glob("/".join([tmp_dirname, "*"]))
            files = set([f.split("/")[-1] for f in files])
            expected = {"Mg_mp-1094122", "MgO_mp-1265", "mol_O2"}
            self.assertEqual(expected, files)

    @unittest.skipIf(not ViseTest.PMG_MAPI_KEY, ViseTest.no_mapi_key)
    def test_get_unary_poscars(self):
        with tempfile.TemporaryDirectory() as tmp_dirname:
            make_poscars_from_mp(elements=["Mg", "O"],
                                 path=tmp_dirname,
                                 only_unary=True)
            files = glob("/".join([tmp_dirname, "*"]))
            files = set([f.split("/")[-1] for f in files])
            expected = {"Mg_mp-1094122", "mol_O2"}
            self.assertEqual(expected, files)

    @patch("vise.util.mp_tools.get_mp_materials")
    def test_get_poscars_with_mock(self, mock):
        with tempfile.TemporaryDirectory() as tmp_dirname:
            make_poscars_from_mp(elements=["Mg", "O"],
                                 path=tmp_dirname)
        mock.assert_called()





# -*- coding: utf-8 -*-

import tempfile
import os
from pathlib import Path

from vise.input_set.input_set import ViseInputSet
from vise.input_set.xc import Xc
from vise.util.testing import ViseTest


class InputSetTest(ViseTest):

    def setUp(self) -> None:
        mgo = self.get_structure_by_name("MgO")
        self.input_set = ViseInputSet.make_input(structure=mgo, xc=Xc.hse)

    def test_write(self):
        with tempfile.TemporaryDirectory() as tmp_dirname:
            os.chdir(tmp_dirname)
            self.input_set.write_input(output_dir=tmp_dirname)
            os.remove("INCAR")
            os.remove("POSCAR")
            os.remove("POTCAR")
            os.remove("KPOINTS")
            os.remove("vise.json")
            Path.cwd()  # may be safer to go back to cwd

    def test_dict(self):
        expected = self.input_set.as_dict()
        actual = ViseInputSet.from_dict(expected).as_dict()
        self.assertEqual(expected, actual)

    def test_msonable(self):
        self.assertMSONable(self.input_set)



# -*- coding: utf-8 -*-
from functools import reduce
from math import pi, log, exp, sqrt
from pathlib import Path
from typing import Union

import yaml
from pymatgen.core.composition import Composition
from scipy.constants import hbar, eV, k, m_u

from vise.util.logger import get_logger
from vise.config import REFERENCE_PRESSURE, ROOM_TEMPERATURE
""" 
Characteristic data is from
McQuarrie and Simon, Physical Chemistry A molecular approach,
"""

logger = get_logger(__name__)
parent = Path(__file__).parent

with open(parent / "molecules" / "molecule_data.yaml", 'r') as fr:
    MOLECULE_DATA = yaml.safe_load(fr)


class Gas:
    def __init__(self,
                 formula: Union[str, Composition],
                 temperature: float = ROOM_TEMPERATURE,
                 pressure: float = REFERENCE_PRESSURE,
                 high_t_limit_factor: float = 20.0):
        """
        Args:
            formula (str/Composition):
                Formula of molecule such as O2 and Composition(O2).
            temperature (float):
                Temperature in K.
            pressure (float):
                Pressure in Pa.
            high_t_limit_factor (float):
                Factor to judge if the given temperature is higher enough than
                the characteristic rotational temperatures. Used whether we
                can use the high-temperature limit approximation.
        """
        self.formula = str(formula)
        if formula not in MOLECULE_DATA:
            raise NotRegisteredError(
                f"{self.formula} is not registered as a molecule in vise."  
                f"Candidates are {', '.join(MOLECULE_DATA.keys())}.")

        self.char_vib_temp = MOLECULE_DATA[formula]["vib"]
        self.char_rot_temp = MOLECULE_DATA[formula]["rot"]
        self.mag_degeneracy = MOLECULE_DATA[formula]["mag_degeneracy"]
        self.sym_num = MOLECULE_DATA[formula]["sym_num"]
        self.is_linear = True if len(self.char_rot_temp) == 1 else False

        self.t = temperature
        self.pressure = pressure

        high_temp_lim_criterion = max(self.char_rot_temp) * high_t_limit_factor
        self.high_temp_lim = high_temp_lim_criterion < self.t

        if self.is_linear is False and self.high_temp_lim is False:
            raise ValueError(
                "Rotational free energy for non-linear molecule is estimated "
                "only at the high-temperature limit. Set more than "
                f"{high_temp_lim_criterion}K")

    @property
    def zero_point_vibrational_energy(self) -> float:
        return sum(t for t in self.char_vib_temp) * k / eV / 2  # eV

    @property
    def kt_in_ev(self) -> float:
        return self.t * k / eV  # [eV]

    @property
    def kt(self) -> float:
        return self.t * k  # [J]

    @property
    def trans_free_energy(self) -> float:
        weight = Composition(self.formula).weight * m_u  # kg
        quantum_volume = (2 * pi * hbar ** 2 / weight / self.kt) ** 1.5  # m^3
        dist_func = self.kt / (self.pressure * quantum_volume)  # -
        return - self.kt_in_ev * log(dist_func)  # eV

    @property
    def rot_free_energy(self) -> float:
        max_sum = 50
        if self.is_linear:
            # high-temperature limit
            if self.high_temp_lim:
                q = self.t / self.char_rot_temp[0]
            else:
                factor = self.char_rot_temp[0] / self.t
                q = sum([(2 * j + 1) * exp(-j * (j + 1) * factor)
                         for j in range(0, max_sum)])
        else:
            # high-temperature limit
            char_temp_mul = reduce(lambda x, y: x*y, self.char_rot_temp)
            q = sqrt(pi * self.t ** 3) / sqrt(char_temp_mul)

        return - self.kt_in_ev * log(q / self.sym_num)

    @property
    def vib_free_energy(self) -> float:
        logs = [log(1 / (1 - exp(-t / self.t))) for t in self.char_vib_temp]
        return - self.kt_in_ev * sum(logs)

    @property
    def spin_free_energy(self) -> float:
        return - self.kt_in_ev * log(self.mag_degeneracy)

    @property
    def n_atoms(self) -> int:
        return Composition(self.formula).num_atoms

    @property
    def chem_pot_shift(self) -> float:
        """Free energy shift (eV/atom) """
        return (self.trans_free_energy + self.rot_free_energy +
                self.vib_free_energy + self.spin_free_energy) / self.n_atoms

    @property
    def energy_shift(self) -> float:
        """Free energy shift including zero point vibration (eV/atom) """
        return (self.chem_pot_shift +
                self.zero_point_vibrational_energy / self.n_atoms)


class NotRegisteredError(Exception):
    pass

